import numpy as np

from ..distributions import Debye, ColeCole, ColeDavidson, KWW, HavriliakNegami
from ..utils.constants import epsilon0


class _AbstractBDS:
    name = 'Abstract'
    type = 'Dielectric Spectroscopy'
    equation = ''
    params = [r'\Delta\epsilon', r'\tau_{0}']
    bounds = [(0, None), (0, None)]
    susceptibility = None
    iscomplex = True

    @classmethod
    def func(cls, x, *args, complex_mode: int = 0, **kwargs):
        # args[0] : Delta epsilon
        # args[1:] : every other parameter
        chi = args[0] * cls.susceptibility(2*np.pi*x, *args[1:], **kwargs)
        if complex_mode == 0:
            return chi
        elif complex_mode == 1:
            return chi.real
        elif complex_mode == 2:
            return chi.imag
        else:
            raise ValueError(f'{complex_mode!r} is not 0, 1, 2')


class DebyeBDS(_AbstractBDS):
    name = 'Debye'
    equation = r'\Delta\epsilon / (1-i\omega\tau)'
    susceptibility = Debye.susceptibility


class ColeColeBDS(_AbstractBDS):
    name = 'Cole-Cole'
    equation = r'\Delta\epsilon / (1-i\omega\tau)^{\alpha}'
    params = _AbstractBDS.params + [r'\alpha']
    bounds = _AbstractBDS.bounds + [(0, 1)]
    susceptibility = ColeCole.susceptibility


class ColeDavidsonBDS(_AbstractBDS):
    name = 'Cole-Davidson'
    equation = r'\Delta\epsilon / [1-(i\omega\tau)^{\gamma})'
    params = _AbstractBDS.params + [r'\gamma']
    bounds = _AbstractBDS.bounds + [(0, 1)]
    susceptibility = ColeDavidson.susceptibility


class HavriliakNegamiBDS(_AbstractBDS):
    name = 'Havriliak-Negami'
    equation = r'\Delta\epsilon / [1-(i\omega\tau)^{\gamma}]^{\alpha}'
    params = _AbstractBDS.params + [r'\alpha', r'\gamma']
    bounds = _AbstractBDS.bounds + [(0, 1), (0, 1)]
    susceptibility = HavriliakNegami.susceptibility


class KWWBDS(_AbstractBDS):
    name = 'KWW'
    params = _AbstractBDS.params + [r'\beta']
    bounds = _AbstractBDS.bounds + [(0, 1)]
    susceptibility = KWW.susceptibility


class EpsInfty:
    name = 'Epsilon Infinity'
    type = 'Dielectric Spectroscopy'
    equation = r'\epsilon_{\infty}'
    params = [r'\epsilon_{\infty}']
    bounds = [(None, None)]
    iscomplex = True

    @staticmethod
    def func(x, eps, complex_mode: int = 0):
        if complex_mode == 0:
            ret_val = np.zeros(x.shape, dtype=complex)
            ret_val += eps
        elif complex_mode == 1:
            ret_val = eps * np.ones(x.shape)
        elif complex_mode == 2:
            ret_val = np.zeros(x.shape)
        else:
            raise ValueError(f'{complex_mode!r} is not 0, 1, 2')

        return ret_val


class CCwithHFW:
    name = 'CC + HF wing'
    type = 'Dielectric Spectroscopy'
    equation = r'\Delta\epsilon CC(\omega, \tau, \alpha) / CD(\omega, \tau_{c}, \beta-\delta)'
    params = [r'\Delta\epsilon', r'\tau', r'\alpha', r'\tau_{c}', '\delta']
    bounds = [(0, None), (0, None), (0, 1), (0, None), (0, 1)]
    iscomplex = True

    @staticmethod
    def func(x, deps, tau, alpha, tauc, delta, complex_mode: int = 0):
        w = 2*np.pi*x
        numerator = (1 + 1j*w*tauc)**(alpha-delta)
        denominator = 1 + (1j*w*tau)**alpha

        epsilon = deps * np.conjugate(numerator / denominator)

        if complex_mode == 0:
            return epsilon
        elif complex_mode == 1:
            return epsilon.real
        elif complex_mode == 2:
            return epsilon.imag
        else:
            raise ValueError(f'{complex_mode!r} has not value 0, 1, or 2')


class CDwithHFW:
    name = 'CD + HF wing'
    type = 'Dielectric Spectroscopy'
    equation = r'\Delta\epsilon CD(\omega, \tau, \gamma) / CD(\omega, \tau_{c}, \gamma-\delta)'
    params = [r'\Delta\epsilon', r'\tau', r'\gamma', r'\tau_{c}', '\delta']
    bounds = [(0, None), (0, None), (0, 1), (0, None), (0, 1)]
    iscomplex = True

    @staticmethod
    def func(x, deps, tau, gamma, tauc, delta, complex_mode: int = 0):
        w = 2*np.pi*x
        numerator = (1 + 1j*w*tauc)**(gamma-delta)
        denominator = (1 + 1j*w*tau)**gamma

        epsilon = deps * np.conjugate(numerator / denominator)

        if complex_mode == 0:
            return epsilon
        elif complex_mode == 1:
            return epsilon.real
        elif complex_mode == 2:
            return epsilon.imag
        else:
            raise ValueError(f'{complex_mode!r} has not value 0, 1, or 2')


class PowerLawBDS:
    name = 'Power Law'
    type = 'Dielectric Spectroscopy'
    equation = r'A / (i\omega)^{n}'
    params = ['A', 'n']
    bounds = [(None, None), (None, None)]
    iscomplex = True

    @staticmethod
    def func(x, a, n, complex_mode: int = 0):
        if complex_mode == 0:
            ret_val = np.exp(1j*n*np.pi/2) * a / x**n
        elif complex_mode == 1:
            ret_val = np.cos(n*np.pi/2) * a / x**n
        elif complex_mode == 2:
            ret_val = np.sin(n*np.pi/2) * a / x**n
        else:
            raise ValueError(f'{complex_mode!r} is not 0, 1, 2')

        return ret_val


class DCCondBDS:
    name = 'DC cond.'
    type = 'Dielectric Spectroscopy'
    equation = r'i\sigma_{dc}/\epsilon_{0}\omega'
    params = [r'\sigma_{dc}']
    bounds = [(0, None)]
    iscomplex = True

    @staticmethod
    def func(x, sigma, complex_mode: int = 0):
        if complex_mode == 0:
            ret_val = np.zeros(x.shape, dtype=complex)
            ret_val += 1j * sigma / x / epsilon0
        elif complex_mode == 1:
            ret_val = np.zeros(x.shape)
        elif complex_mode == 2:
            ret_val = sigma / x / epsilon0
        else:
            raise ValueError(f'{complex_mode!r} is not 0, 1, 2')

        return ret_val


class DerivativeHavriliakNegami:
    name = 'Derivative HN'
    type = 'Dielectric Spectroscopy'
    params = [r'\Delta\epsilon', r'\tau', r'\alpha', r'\gamma']
    bounds = [(0, None), (0, None), (0, 1), (0, 1)]

    @staticmethod
    def func(x, eps, tau, a, g):
        omtau = 2*np.pi*x * tau
        theta = np.arctan2(np.sin(np.pi*a/2.), omtau**(-a) + np.cos(np.pi*a/2.))

        numer = a*g * omtau**a * np.cos(np.pi*a/2. - (1+g)*theta)
        denom = (1 + 2*omtau**a * np.cos(np.pi*a/2.) + omtau**(2*a))**((1+g)/2.)

        return eps * np.pi * numer / denom / 2.


class DerivativeColeCole:
    name = 'Derivative CC'
    type = 'Dielectric Spectroscopy'
    params = [r'\Delta\epsilon', r'\tau', r'\alpha']
    bounds = [(0, None), (0, None), (0, 1)]

    @staticmethod
    def func(x, eps, tau, alpha):
        omtau = 2*np.pi*x * tau
        theta = np.arctan2(np.sin(np.pi*alpha/2.), omtau**(-alpha) + np.cos(np.pi*alpha/2.))

        numer = alpha * omtau**alpha * np.cos(np.pi*alpha/2. - 2*theta)
        denom = 1 + 2*omtau**alpha * np.cos(np.pi*alpha/2.) + omtau**(2*alpha)

        return eps * np.pi * numer / denom / 2.


class DerivativeColeDavidson:
    name = 'Derivative CD'
    type = 'Dielectric Spectroscopy'
    params = [r'\Delta\epsilon', r'\tau', r'\gamma']
    bounds = [(0, None), (0, None), (0, 1)]

    @staticmethod
    def func(x, eps, tau, g):
        omtau = 2*np.pi*x * tau
        numer = g * omtau * np.sin((1+g)*np.arctan(omtau))
        denom = (1 + omtau**2)**((1+g)/2.)

        return eps * np.pi * numer / denom / 2.
